Option Explicit On 
Option Strict On
Option Compare Binary

Imports System

' <remarks>
'   Stellt Methoden zum Berechnen des Binomialkoeffizienten 
'   n ber k bereit.
' </remarks>
Public Class BinomCoeff

    ' <summary>
    '   Gibt <paramref name="n"/> ber <paramref name="k"/> zurck. 
    '   Die Berechnung erfolgt ber ein iteratives Verfahren.
    ' </summary>
    '
    ' Berechnet n ber k nach der Formel 
    '    B(n, k) = n! / k! * (n - k)!.
    '
    Public Shared Function BiFac( _
      ByVal n As Integer, _
      ByVal k As Integer) As Integer
        If n < k Then
            Throw New Exception("n must not be smaller than k.")
        Else
            Return FacIterative(n) \ (FacIterative(k) * (FacIterative(n - k)))
        End If
    End Function

    Private Shared Function FacIterative( _
      ByVal i As Integer) As Integer
        Dim j As Integer, k As Integer = 1
        For j = 1 To i
            k = k * j
        Next j
        Return k
    End Function

    ' <summary>
    '   Gibt <paramref name="n"/> ber <paramref name="k"/> zurck. 
    '   Die Berechnung erfolgt ber ein iteratives Verfahren.
    ' </summary>
    '
    ' Definition: B(n, k) = n! / (n - k)! * k!.
    ' Obige Formel kann gekrzt werden, wenn k < n ist, was immer zutrifft.
    '
    Public Shared Function BinomialCoeff( _
      ByVal n As Integer, _
      ByVal k As Integer) As Integer
        If n < k Then
            Throw New Exception("n must not be smaller than k.")
        Else
            Dim i As Integer
            Dim n1 As Integer = 1
            For i = (n - k + 1) To n
                n1 = n1 * i
            Next i
            Dim n2 As Integer = 1
            For i = 1 To k
                n2 = n2 * i
            Next i
            Return CInt(n1 / n2)
        End If
    End Function

    ' <summary>
    '   Gibt <paramref name="n"/> ber <paramref name="k"/> zurck. 
    '   Die Berechnung erfolgt ber ein rekursives Verfahren.
    ' </summary>
    '
    ' Berechnen des Binomialkoeffizienten ber das herkmmliche 
    ' rekursive Verfahren. Also ber die Rekursionsformel
    '   B(n, k) = B(n - 1, k - 1) + B(n - 1, k).
    ' Dieses Verfahren ist wegen seiner exponentiellen Laufzeit 
    ' fr grssere Instanzen nicht verwendbar.
    '
    Public Shared Function BiRecursive( _
      ByVal n As Integer, _
      ByVal k As Integer) As Integer
        If n < k Then
            Throw New Exception("n must not be smaller than k.")
        Else
            Return BiRec(n, k)
        End If
    End Function

    Private Shared Function BiRec( _
      ByVal n As Integer, _
      ByVal k As Integer) As Integer
        If k = 0 Or n = k Then
            Return 1
        Else
            Return BiRec(n - 1, k - 1) + BiRec(n - 1, k)
        End If
    End Function

    ' <summary>
    '   Gibt <paramref name="n"/> ber <paramref name="k"/> zurck. 
    '   Die Berechnung erfolgt ber ein Verfahren, das auf 
    '   dynamischer Programmierung beruht.
    ' </summary>
    '
    ' Berechnung des Binomialkoeffizienten ber einen Algorithmus, 
    ' der auf dem Prinzip der dynamischen Programmierung beruht.
    ' Dadurch kann die Laufzeit wesentlich verringert werden. 
    ' Durch Ausntzen der Tatsache, dass n ber k gleich n ber n - k ist,
    ' kann man den Aufwand zur Berechnung noch einmal reduzieren.
    '
    ' Die Idee zu diesem Algorithmus basiert auf dem Aufbau 
    ' des Pascalschen Dreiecks, das wie folgt aussieht:
    '
    '       1
    '      1 1
    '     1 2 1
    '    1 3 3 1
    '   1 4 6 4 1
    '
    ' Der Wert von n ber k lsst sich dabei leicht ablesen, 
    ' indem man n "Ebenen" nach unten geht und dann k
    ' "Ebenen" nach rechts. Diese Pyramide wird im Datenfeld B 
    ' abgebildet, es werden allerdings nur die fr die
    ' Berechnung erforderlichen Werte der Pyramide berechnet. 
    ' Die Pyramide ist fr n ber k, n > k immer n Stufen
    ' hoch und muss auf k Elemente "Breite" berechnet werden. 
    ' Diese Version des Algorithmus berechnet wirklich nur
    ' genau jene Felder, die man bentigt, es werden keine 
    ' Doppelberechnungen durchgefhrt und die Matrix braucht
    ' nicht initialisiert zu werden. Laufzeit ist Theta((n - k) * k). 
    ' Ein weiterer Vorteeil des Verfahrens ist, dass
    ' keine Multiplikationen und Divisionen erforderlich sind.
    '
    Public Shared Function BiDynprog( _
      ByVal n As Integer, _
      ByVal k As Integer) As Integer
        If n < k Then
            Throw New Exception("n must not be smaller than k.")
        Else
            If n - k < k Then
                k = n - k
            End If
            Dim B(n, k) As Integer
            GoUpLeft(B, n, k)
            Return B(n, k)
        End If
    End Function

    Private Shared Sub GoUpLeft( _
      ByRef B(,) As Integer, _
      ByVal n As Integer, _
      ByVal k As Integer)

        ' Es wurde der obere Rand der Matrix erreicht oder 
        ' wir befinden uns auf der Diagonale. In diesem Fall
        ' setzen wir die Rekursion nicht fort, sondern beenden 
        ' sie und geben 1 zurck. Der obere Rand entspricht
        ' der rechten "Seite" des Pascalschen Dreiecks, der linke 
        'Rand der Diagonale. Man muss sich also oben
        ' angegebene Zeichnung um 90 Grad nach rechts rotiert vorstellen.
        If k = 0 Or k = n Then
            B(n, k) = 1

        ' Wir befinden uns "im Feld" und setzen die Rekursion fort.
        Else
            GoUpLeft(B, n - 1, k - 1)
            If k = 1 Then
                GoUpLeft(B, n - 1, k)
            Else
                GoLeft(B, n - 1, k)
            End If
            B(n, k) = B(n - 1, k - 1) + B(n - 1, k)
        End If
    End Sub

    Private Shared Sub GoLeft( _
      ByRef B(,) As Integer, _
      ByVal n As Integer, _
      ByVal k As Integer)
        If k = n Then
            B(n, k) = 1
        Else
            GoLeft(B, n - 1, k)
            B(n, k) = B(n - 1, k - 1) + B(n - 1, k)
        End If
    End Sub
End Class